@@ -1,23 +1,18 @@
-from __future__ import absolute_import, division, print_function
-from builtins import object, str
-import re
-
-# import for either Python 2 or 3
-try:
-    from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit
-except ImportError:
-    from urllib.parse import parse_qs, urlsplit, urlunsplit
-    from urllib.parse import urlencode
-
 import ipaddress
+import json
+import os
+import re
+import time
 from functools import lru_cache
+from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit
 
 import requests
 
 import rdap
+import rdap.bootstrap
 from rdap.config import Config
-from rdap.objects import RdapAsn, RdapObject, RdapNetwork, RdapDomain, RdapEntity
 from rdap.exceptions import RdapHTTPError, RdapNotFoundError
+from rdap.objects import RdapAsn, RdapDomain, RdapEntity, RdapNetwork
 
 
 class RdapRequestAuth(requests.auth.AuthBase):
@@ -70,21 +65,45 @@ def strip_auth(url):
     return url
 
 
-class RdapClient(object):
+class RdapClient:
     """
-    Client to do RDAP queries against defined bootstrap URL.
+    Client to do RDAP queries.
     """
 
-    def __init__(self, config=None):
+    def __init__(self, config=None, config_dir=None):
         """
+        Initialize an RdapClient.
+
         config is a dict or rdap.config.Config object
+        config_dir is a string pointing to a config directory
         """
-        if not config:
+        if config:
+            if config_dir:
+                raise ValueError("Cannot pass config and config_dir")
+
+            if isinstance(config, Config):
+                self.config_dir = config.meta.get("config_dir", None)
+                config = config.get("rdap")
+            else:
+                self.config_dir = None
+
+        elif config_dir:
+            config = Config(read=config_dir).get("rdap")
+            self.config_dir = config_dir
+
+        else:
+            self.config_dir = None
             config = dict()
-        elif isinstance(config, Config):
-            config = config.get("rdap")
+
+        self.config = config
 
         self.url = config.get("bootstrap_url", "https://rdap.org/").rstrip("/")
+        self.self_bootstrap = config.get("self_bootstrap")
+
+        # check for explicit bootstrap dir, otherwise set it to config_dir/bootstrap
+        self.bootstrap_dir = config.get("bootstrap_dir")
+        if not self.bootstrap_dir and self.config_dir:
+            self.bootstrap_dir = os.path.join(self.config_dir, "bootstrap")
 
         # use setter
         self._recurse_roles = None
@@ -94,11 +113,14 @@ class RdapClient(object):
 
         self._asn_req = None
         self._history = []
-        self.timeout = config.get("timeout", 0.5)
+        self._asn_tree = None
+        self.timeout = config.get("timeout")
 
         self.http = requests.Session()
         self.http.auth = RdapRequestAuth(
-            **dict(lacnic_apikey=config.get("lacnic_apikey", None),)
+            **dict(
+                lacnic_apikey=config.get("lacnic_apikey", None),
+            )
         )
 
         self.http.headers["User-Agent"] = "20C-rdap/{} {}".format(
@@ -121,6 +143,65 @@ class RdapClient(object):
             raise RdapNotFoundError(msg)
         raise RdapHTTPError(msg)
 
+    def asn_url(self, asn):
+        """Gets the correct url for specified ASN."""
+        if not self.self_bootstrap:
+            return self.url
+
+        if not self._asn_tree:
+            self._asn_tree = rdap.bootstrap.AsnTree(self.get_bootstrap_data("asn"))
+        return self._asn_tree.get_service(asn).url
+
+    def fetch_bootstrap_data(self, typ):
+        "Fetches bootstrap data in json."
+        # TODO - check modified header
+        return self._get(
+            self.config.get("bootstrap_data_url", "https://data.iana.org/rdap/")
+            + f"{typ}.json"
+        ).json()
+
+    def get_bootstrap_data(self, typ):
+        """Checks for a local cache, fetches bootstrap data and returns it."""
+
+        try:
+            data_file = os.path.join(self.bootstrap_dir, f"{typ}.json")
+            cache_age = time.time() - self.config.get("bootstrap_cache_ttl") * 3600
+            if os.path.getmtime(data_file) > cache_age:
+                with open(data_file) as fh:
+                    return json.load(fh)
+
+        except (FileNotFoundError, TypeError):
+            pass
+
+        data = self.fetch_bootstrap_data(typ)
+
+        # no cache if there's no bootstrap_dir
+        if not self.bootstrap_dir:
+            return data
+
+        if not os.path.isdir(self.bootstrap_dir):
+            os.mkdir(self.bootstrap_dir)
+
+        with open(data_file, "w") as fh:
+            json.dump(data, fh, separators=(",", ":"))
+        return data
+
+    def write_bootstrap_data(self, typ):
+        """Fetches and writes bootstrap data."""
+
+        if not self.bootstrap_dir:
+            raise ValueError("bootstrap dir is not set")
+
+        data = self.fetch_bootstrap_data(typ)
+
+        if not os.path.isdir(self.bootstrap_dir):
+            os.mkdir(self.bootstrap_dir)
+
+        data_file = os.path.join(self.bootstrap_dir, f"{typ}.json")
+        print(f"writing {data_file}")
+        with open(data_file, "w") as fh:
+            json.dump(data, fh, separators=(",", ":"))
+
     @property
     def history(self):
         return self._history
@@ -161,7 +242,7 @@ class RdapClient(object):
         if "." in qstr:
             return self.get_domain(qstr)
 
-        raise NotImplementedError("unknown query {}".format(query))
+        raise NotImplementedError(f"unknown query {query}")
 
     @lru_cache(maxsize=1024)
     def get_rdap(self, url):
@@ -181,20 +262,21 @@ class RdapClient(object):
         classname = data.get("objectClassName", None)
         if not classname:
             raise NotImplementedError(
-                "query '{}' did not return an objectClassName".format(url)
+                f"query '{url}' did not return an objectClassName"
             )
         if classname in classes:
             return classes[classname](data, self)
         else:
             raise NotImplementedError(
-                "Unknown objectClassName '{}' from '{}'".format(classname, url)
+                f"Unknown objectClassName '{classname}' from '{url}'"
             )
 
     def get_asn(self, asn):
         """
         Get an ASN object.
         """
-        url = "{}/autnum/{}".format(self.url, int(asn))
+        asn = int(asn)
+        url = "{}/autnum/{}".format(self.asn_url(asn), asn)
         # save reqest to get url for following entity lookups
         self._asn_req = self._get(url)
         return RdapAsn(self._asn_req.json(), self)
@@ -203,21 +285,21 @@ class RdapClient(object):
         """
         Get a domain object.
         """
-        url = "{}/domain/{}".format(self.url, domain)
+        url = f"{self.url}/domain/{domain}"
         return RdapAsn(self._get(url).json(), self)
 
     def get_ip(self, address):
         """
         Get an IP object.
         """
-        url = "{}/ip/{}".format(self.url, address)
+        url = f"{self.url}/ip/{address}"
         return RdapNetwork(self._get(url).json(), self)
 
     def get_entity(self, handle, base_url):
         """
         get entity information in object form
         """
-        url = "{}/entity/{}".format(base_url, handle)
+        url = f"{base_url}/entity/{handle}"
         return RdapEntity(self._get(url).json(), self)
 
     def get_entity_url(self, handle):
@@ -231,7 +313,7 @@ class RdapClient(object):
         else:
             url = self.url
 
-        url = "{}/entity/{}".format(url, handle)
+        url = f"{url}/entity/{handle}"
         return url
 
     def get_data(self, url):
